/*-
 * Copyright (C)2000..2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2000..2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: pass3.c,v 1.17 2014/09/08 17:10:53 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3

#include <sys/types.h>
#include <sysexits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <babolo/BLINflag.h>
#include "getCGIparm_int.h"

static int
/*****************************************************************************************************/
quote_literal(const char *pattern, const char *query, char **result) {                           /****
 *****************************************************************************************************/
    const char *q;
    char *p;
    int i;

    i = 0;
    q = query;
    *result = malloc(strlen(query) * 2 + 1);
    if  (!*result) return(EX_TEMPFAIL);
    p = *result;
    for (; *q; q++) {
        switch(*q) {
        case '\'': case '\\': i |= 1; *p++ = *q; *p++ = *q; break;
        default: *p++ = *q;
    }   }
    *p++ = '\0';
    return(EX_OK);
};

static int
/*****************************************************************************************************/
quote_ident(const char *pattern, const char *query, char **result) {                             /****
 *****************************************************************************************************/
    const char *q;
    char *p;
    int i;

    i = 0;
    q = query;
    *result = malloc(strlen(query) * 2 + 1);
    if  (!*result) return(EX_TEMPFAIL);
    p = *result;
    for (; *q; q++) {
        switch(*q) {
        case '\"': case '\\': i |= 1; *p++ = *q; *p++ = *q; break;
        default: *p++ = *q;
    }   }
    *p++ = '\0';
    return(EX_OK);
};

static int
/*****************************************************************************************************/
quote_copy(const char *pattern, const char *query, char **result) {                              /****
 *****************************************************************************************************/
    const char *q;
    char *p;
    int i;

    q = query;
    i = 0;
    *result = malloc(strlen(query) * 2 + 1);
    if  (!*result) return(EX_TEMPFAIL);
    p = *result;
    for (; *q; q++) {
        switch(*q) {
        case 0x08: i |= 1; *p++ = '\\'; *p++ = 'b';   break;
        case 0x09: i |= 1; *p++ = '\\'; *p++ = 't';   break;
        case 0x0A: i |= 1; *p++ = '\\'; *p++ = 'n';   break;
        case 0x0B: i |= 1; *p++ = '\\'; *p++ = 'v';   break;
        case 0x0C: i |= 1; *p++ = '\\'; *p++ = 'f';   break;
        case 0x0D: i |= 1; *p++ = '\\'; *p++ = 'r';   break;
        case '\\': i |= 1; *p++ = '\\'; *p++ = '\\';  break;
        default: *p++ = *q;
    }   }
    *p++ = '\0';
    return(EX_OK);
};

static int
/*****************************************************************************************************/
quote_single(const char *pattern, const char *query, char **result) {                            /****
 *****************************************************************************************************/
    char *p;

    *result = malloc(strlen(query) + 3);
    if  (!*result) return(EX_TEMPFAIL);
    p = *result;
    strcpy(p, "\'");
    strcat(p, query);
    strcat(p, "\'");
    return(EX_OK);
};

static int
/*****************************************************************************************************/
quote_double(const char *pattern, const char *query, char **result) {                            /****
 *****************************************************************************************************/
    char *p;

    *result = malloc(strlen(query) + 3);
    if  (!*result) return(EX_TEMPFAIL);
    p = *result;
    strcpy(p, "\"");
    strcat(p, query);
    strcat(p, "\"");
    return(EX_OK);
};

const getCGIparmquotens_t
/*****************************************************************************************************/
getCGIparmquotens[] =                                                                            /****
 *****************************************************************************************************/
{ {"psql" , 0, quote_literal}
, {"psqi" , 0, quote_ident  }
, {"copy" , 0, quote_copy   }
, {"sq"   , 0, quote_single }
, {"dq"   , 0, quote_double }
, {NULL   , 0, NULL         }
};

int
/*****************************************************************************************************
 *****************************************************************************************************/
getCGIparmpass3(BLIN_flag flags, const u_char *qq, u_char **unesc) {                             /****
 *****************************************************************************************************
 *****************************************************************************************************/
#   define unescaped ((char**)unesc)
    int i, ex;
    char *r;
    char *q;

#   define blin_internal_flags (flags & BLIN_MASK)
    q = qq ? strdup((char*)qq) : NULL;
    ifBLIN_QX2( "flags=%06X, q=%s~, unescaped=%s~"
              , flags
              , qq ? (char*)qq : ""
              , (unescaped && *unescaped) ? *unescaped : ""
              );
    ex = EX_OK;
    if  (!unescaped) {
        ex = EX_SOFTWARE;
    } else if (!q || !*q) {
        ex = EX_OK;
    } else if (~flags & gCGI_ZERO || **unescaped) {
        char *p = NULL, *f1 = NULL;

        while (!ex && q && *q) {
            ex = EX_USAGE;
            r = strsep(&q, "/");
            for (i = 0; getCGIparmquotens[i].name; i++) {
                if  ( getCGIparmquotens[i].namelen
                    ? !strncmp(getCGIparmquotens[i].name, r, getCGIparmquotens[i].namelen)
                    : !strcmp(getCGIparmquotens[i].name, r)
                    ) {
                    ifBLIN_QX2("r=%s unesc=%s", r, *unescaped);
                    ex = getCGIparmquotens[i].test_proc(r, *unescaped, &p) ? EX_DATAERR : EX_OK;
                    if  (ex) goto out;
                    if  (*unescaped != p) {
                        if  (f1 && *unescaped == f1) free(*unescaped);
                        *unescaped = f1 = p;
                        ifBLIN_QX1("%d NEW=%s~", ex, p);
                    } else {
                        ifBLIN_QX1("%d OLD=%s~", ex, p);
                    }
                    break;
    }   }   }   }
out:
    free(q);
    ifBLIN_QX2("ex=%d", ex);
    return(ex);
#   undef blin_internal_flags
}
